package dev.fitko.fitconnect.client;

import dev.fitko.fitconnect.api.domain.model.destination.Destination;
import dev.fitko.fitconnect.api.domain.model.event.Event;
import dev.fitko.fitconnect.api.domain.model.event.EventPayload;
import dev.fitko.fitconnect.api.domain.model.event.SubmissionStatus;
import dev.fitko.fitconnect.api.domain.model.event.problems.Problem;
import dev.fitko.fitconnect.api.domain.model.metadata.Metadata;
import dev.fitko.fitconnect.api.domain.model.metadata.attachment.ApiAttachment;
import dev.fitko.fitconnect.api.domain.model.metadata.data.Data;
import dev.fitko.fitconnect.api.domain.model.submission.SentSubmission;
import dev.fitko.fitconnect.api.domain.model.submission.Submission;
import dev.fitko.fitconnect.api.domain.model.submission.SubmissionForPickup;
import dev.fitko.fitconnect.api.domain.validation.ValidationResult;
import dev.fitko.fitconnect.api.exceptions.client.FitConnectSubscriberException;
import dev.fitko.fitconnect.api.services.Subscriber;
import dev.fitko.fitconnect.client.sender.model.Attachment;
import dev.fitko.fitconnect.client.subscriber.ReceivedSubmission;
import dev.fitko.fitconnect.client.subscriber.SubmissionReceiver;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.List;
import java.util.Set;
import java.util.UUID;
import java.util.function.Supplier;

/**
 * A subscriber-side client for retrieving submissions.
 */
public class SubscriberClient {

    private static final Logger LOGGER = LoggerFactory.getLogger(SubscriberClient.class);

    private static final int DEFAULT_SUBMISSION_LIMIT = 500;

    private final Subscriber subscriber;
    private final SubmissionReceiver submissionReceiver;

    public SubscriberClient(final Subscriber subscriber, final SubmissionReceiver submissionReceiver) {
        this.subscriber = subscriber;
        this.submissionReceiver = submissionReceiver;
    }

    /**
     * Loads a list of available {@link SubmissionForPickup} that were submitted to the {@link Subscriber}s destination.
     *
     * @param destinationId filter criterion for the submissions
     * @return set of the first 0..500 available submissions
     * @throws FitConnectSubscriberException if a technical error occurred retrieving the list of available submissions
     */
    public Set<SubmissionForPickup> getAvailableSubmissionsForDestination(final UUID destinationId) throws FitConnectSubscriberException {
        return getAvailableSubmissionsForDestination(destinationId, 0, DEFAULT_SUBMISSION_LIMIT);
    }

    /**
     * Loads a list of available {@link SubmissionForPickup} that were submitted to the {@link Subscriber}s destination.
     *
     * @param destinationId filter criterion for the submissions
     * @param offset        position in the dataset
     * @param limit         number of submissions in result (max. is 500)
     * @return set of available submissions in the given range
     * @throws FitConnectSubscriberException if a technical error occurred retrieving the list of available submissions
     */
    public Set<SubmissionForPickup> getAvailableSubmissionsForDestination(final UUID destinationId, final int offset, final int limit) throws FitConnectSubscriberException {
        final Set<SubmissionForPickup> submissions = wrapExceptions(() -> subscriber.pollAvailableSubmissionsForDestination(destinationId, offset, limit), "Submissions could not be retrieved");
        LOGGER.info("Received {} submission(s) for destination {}", submissions.size(), destinationId);
        return submissions;
    }

    /**
     * Loads a single {@link SubmissionForPickup} by id.
     *
     * @param submission {@link SentSubmission} that is requested
     * @return {@link ReceivedSubmission} to get the submission {@link Metadata}, {@link Data} and {@link ApiAttachment}s as well as
     * accept or reject the loaded submission
     * @throws FitConnectSubscriberException if a technical error occurred or validation failed
     */
    public ReceivedSubmission requestSubmission(final SubmissionForPickup submission) throws FitConnectSubscriberException {
        return requestSubmission(submission.getSubmissionId());
    }

    /**
     * Loads a single {@link SubmissionForPickup} by id. Auto-rejects invalid submissions where e.g. a validation failed.
     *
     * @param submissionId unique identifier of the requested submission
     * @return {@link ReceivedSubmission} to get the submission {@link Metadata}, {@link Data} and {@link Attachment}s as well as accept or reject the loaded submission
     * @throws FitConnectSubscriberException if a technical error occurred or validation failed
     */
    public ReceivedSubmission requestSubmission(final UUID submissionId) throws FitConnectSubscriberException {
        return wrapExceptions(() -> submissionReceiver.receiveSubmission(submissionId), "Submission could not be retrieved");
    }

    /**
     * Send a reject-submission {@link Event} to confirm that the handed in submission was rejected.
     *
     * @param submissionForPickup the submission that should be rejected
     * @param rejectionProblems   list of {@link Problem}s that give more detailed information on why the submission was rejected
     * @throws FitConnectSubscriberException if a technical error occurred rejecting the submission
     * @see <a href="https://docs.fitko.de/fit-connect/docs/getting-started/event-log/events">FIT-Connect Events</a>
     */
    public void rejectSubmission(final SubmissionForPickup submissionForPickup, final List<Problem> rejectionProblems) throws FitConnectSubscriberException {
        wrapExceptions(() -> {
            subscriber.rejectSubmission(EventPayload.forRejectEvent(submissionForPickup, rejectionProblems));
            return null;
        }, "Could not reject submission");
    }

    /**
     * Checks if a received callback can be trusted by validating the provided request data
     *
     * @param hmac           authentication code provided by the callback
     * @param timestampInSec timestamp in seconds provided by the callback
     * @param httpBody       HTTP body provided by the callback
     * @param callbackSecret secret owned by the client, which is used to calculate the hmac
     * @return {@code true} if hmac and timestamp provided by the callback meet the required conditions
     * @throws FitConnectSubscriberException if a technical error occurred during the validation
     */
    public ValidationResult validateCallback(final String hmac, final Long timestampInSec, final String httpBody, final String callbackSecret) throws FitConnectSubscriberException {
        return wrapExceptions(() -> subscriber.validateCallback(hmac, timestampInSec, httpBody, callbackSecret), "Callback validation failed");
    }

    /**
     * Retrieve the current status of a {@link Submission}.
     *
     * @param destinationId unique identifier of the {@link Destination} the status should be retrieved for
     * @param caseId        unique identifier of the case the status should be retrieved for
     * @param submissionId  unique identifier of the submission the status should be retrieved for
     * @return {@link SubmissionStatus} the current status
     * @throws FitConnectSubscriberException if a technical error occurred during the status request
     */
    public SubmissionStatus getStatusForSubmission(final UUID destinationId, final UUID caseId, final UUID submissionId) throws FitConnectSubscriberException {
        return wrapExceptions(() -> subscriber.getLastedEvent(destinationId, caseId, submissionId), "Status for submission not available");
    }

    private <T> T wrapExceptions(final Supplier<T> supplier, final String errorMessage) {
        try {
            return supplier.get();
        } catch (final Exception e) {
            throw new FitConnectSubscriberException(errorMessage, e.getCause() != null ? e.getCause() : e);
        }
    }
}
