package dev.fitko.fitconnect.client.bootstrap;

import dev.fitko.fitconnect.api.config.ApplicationConfig;
import dev.fitko.fitconnect.api.config.Environment;
import dev.fitko.fitconnect.api.config.EnvironmentName;
import dev.fitko.fitconnect.api.config.SenderConfig;
import dev.fitko.fitconnect.api.config.defaults.DefaultEnvironments;
import dev.fitko.fitconnect.api.config.defaults.ResourcePaths;
import dev.fitko.fitconnect.api.exceptions.client.FitConnectInitialisationException;
import org.junit.jupiter.api.Assertions;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.api.io.TempDir;

import java.io.File;
import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.util.List;
import java.util.Map;
import java.util.Objects;

import static dev.fitko.fitconnect.api.config.defaults.DefaultEnvironments.PROD;
import static dev.fitko.fitconnect.api.config.defaults.DefaultEnvironments.STAGE;
import static dev.fitko.fitconnect.api.config.defaults.DefaultEnvironments.TEST;
import static org.hamcrest.MatcherAssert.assertThat;
import static org.hamcrest.Matchers.containsString;
import static org.hamcrest.Matchers.equalTo;
import static org.hamcrest.Matchers.hasSize;
import static org.hamcrest.Matchers.is;
import static org.junit.jupiter.api.Assertions.assertEquals;
import static org.junit.jupiter.api.Assertions.assertFalse;
import static org.junit.jupiter.api.Assertions.assertNotNull;
import static org.junit.jupiter.api.Assertions.assertTrue;

class ApplicationConfigLoaderTest {

    @Test
    void testLoadMissingConfigFile() {
        Assertions.assertThrows(FitConnectInitialisationException.class, () -> ApplicationConfigLoader.loadConfigFromPath(Path.of("/no/path")));
    }

    @Test
    void testEnvironmentNotAvailable() throws IOException {

        // Given
        final String testConfigYaml = getResourceAsString("test-config-with-missing-env.yml");
        final ApplicationConfig config = ApplicationConfigLoader.loadConfigFromYamlString(testConfigYaml);

        //When
        final FitConnectInitialisationException exception = Assertions.assertThrowsExactly(FitConnectInitialisationException.class, config::getCurrentEnvironment);

        //Then
        assertThat(exception.getMessage(), containsString("No environment with name 'ENVIRONMENT_THAT_DOES_NOT_EXIST' found"));
    }

    @Test
    void testLoadTestConfigFromString() throws IOException {

        // Given
        final String testConfigYaml = getResourceAsString("test-config.yml");

        final Environment devEnv = new Environment(
                "https://auth-testing.fit-connect.fitko.dev",
                "https://routing-api-testing.fit-connect.fitko.dev",
                List.of("https://submission-api-testing.fit-connect.fitko.dev"),
                "https://portal.auth-testing.fit-connect.fitko.dev",
                true,
                false,
                false
        );

        final SenderConfig senderConfig = new SenderConfig("1", "123");

        // When
        final ApplicationConfig testConfig = ApplicationConfigLoader.loadConfigFromYamlString(testConfigYaml);

        // Then
        assertNotNull(testConfig);
        assertThat(testConfig.getCurrentEnvironment(), equalTo(devEnv));
        assertThat(testConfig.getSenderConfig(), equalTo(senderConfig));
    }

    @Test
    void testLoadTestConfigFromPath(@TempDir final Path tempDir) throws IOException {

        // Given
        final String testConfigYaml = getResourceAsString("test-config.yml");

        final Path configPath = Files.writeString(Path.of(tempDir.toString(), "test-config.yml"), testConfigYaml);

        // When
        final ApplicationConfig testConfig = ApplicationConfigLoader.loadConfigFromPath(configPath);

        // Then
        assertNotNull(testConfig);
        assertThat(testConfig.getActiveEnvironment(), is(new EnvironmentName("TEST")));
    }

    @Test
    void testSubscriberConfigContainsListOfPrivateDecryptionKeys(@TempDir final Path tempDir) throws IOException {

        // Given
        final String testConfigYaml = getResourceAsString("test-config.yml");
        final Path configPath = Files.writeString(Path.of(tempDir.toString(), "test-config.yml"), testConfigYaml);

        // When
        final ApplicationConfig testConfig = ApplicationConfigLoader.loadConfigFromPath(configPath);

        // Then
        assertNotNull(testConfig);
        assertThat(testConfig.getSubscriberConfig().getPrivateDecryptionKeyPaths(), hasSize(2));
    }

    @Test
    void testUseDefaultEnvironment(@TempDir final Path tempDir) throws IOException {

        // Given
        final String testConfigYaml = getResourceAsString("test-config-default.yml");
        final Path configPath = Files.writeString(Path.of(tempDir.toString(), "test-config-default.yml"), testConfigYaml);

        // When
        final ApplicationConfig config = ApplicationConfigLoader.loadConfigFromPath(configPath);

        // Then
        assertEquals(DefaultEnvironments.getEnvironmentsAsMap().size(), config.getEnvironments().size());

        assertThat(config.getEnvironments().get(PROD.getEnvironmentName()), is(PROD.getEnvironment()));
        assertThat(config.getEnvironments().get(STAGE.getEnvironmentName()), is(STAGE.getEnvironment()));
        assertThat(config.getEnvironments().get(TEST.getEnvironmentName()), is(TEST.getEnvironment()));

    }

    @Test
    void testUseDefaultEnvironmentWithProxyConfig(@TempDir final Path tempDir) throws IOException {

        // Given
        final String testConfigYaml = getResourceAsString("test-config-default.yml");
        final Path configPath = Files.writeString(Path.of(tempDir.toString(), "test-config-default.yml"), testConfigYaml);

        // When
        final ApplicationConfig config = ApplicationConfigLoader.loadConfigFromPath(configPath);

        // Then
        assertTrue(config.getProxyConfig().isProxySet());
        assertThat(config.getProxyConfig().getHttpProxyHost(), is("https://proxy.test.net"));
        assertThat(config.getProxyConfig().getHttpProxyPort(), is(8080));
        assertThat(config.getProxyConfig().getRequestTimeoutInSeconds(), is(10));


    }

    @Test
    void testOverridePropertiesOnDefaultEnvironment(@TempDir final Path tempDir) throws IOException {

        // Given
        final String testConfigYaml = getResourceAsString("test-config.yml");
        final Path configPath = Files.writeString(Path.of(tempDir.toString(), "test-config.yml"), testConfigYaml);

        assertTrue(TEST.getEnvironment().getAllowInsecurePublicKey());

        // When
        final ApplicationConfig config = ApplicationConfigLoader.loadConfigFromPath(configPath);
        final Environment envWithChangedDefaults = config.getEnvironments().get(TEST.getEnvironmentName());

        // Then
        assertFalse(envWithChangedDefaults.getAllowInsecurePublicKey());
    }

    @Test
    void testLoadingAdditionalCustomEnvironment(@TempDir final Path tempDir) throws IOException {

        // Given
        final String testConfigYaml = getResourceAsString("test-config-with-custom-env.yml");
        final Path configPath = Files.writeString(Path.of(tempDir.toString(), "test-config-with-custom-env.yml"), testConfigYaml);

        final Environment customEnvironment = new Environment();
        customEnvironment.setEnableAutoReject(true);
        customEnvironment.setAllowInsecurePublicKey(false);
        customEnvironment.setSkipSubmissionDataValidation(false);
        customEnvironment.setAuthBaseUrl("https://auth-url.test.net");
        customEnvironment.setRoutingBaseUrl("https://routing-url.test.net");
        customEnvironment.setSubmissionBaseUrls(List.of("https://submission-url-1.test.net", "https://submission-url-2.test.net"));
        customEnvironment.setSelfServicePortalBaseUrl("https://portal-url.test.net");

        // When
        final ApplicationConfig configWithCustomEnvironment = ApplicationConfigLoader.loadConfigFromPath(configPath);

        // Then
        assertNotNull(configWithCustomEnvironment);

        final Map<EnvironmentName, Environment> environments = configWithCustomEnvironment.getEnvironments();

        assertThat(environments.size(), is(DefaultEnvironments.getEnvironmentsAsMap().size() + 1));
        assertThat(environments.get(new EnvironmentName("CUSTOM_ENVIRONMENT")), is(customEnvironment));

        // currently there is only one url supported
        assertThat(configWithCustomEnvironment.getSubmissionEndpoint(), is("https://submission-url-1.test.net" + ResourcePaths.SUBMISSION_PATH));

        assertThat(environments.get(PROD.getEnvironmentName()), is(PROD.getEnvironment()));
        assertThat(environments.get(STAGE.getEnvironmentName()), is(STAGE.getEnvironment()));
        assertThat(environments.get(TEST.getEnvironmentName()), is(TEST.getEnvironment()));
    }

    @Test
    void testLocalSubmissionDataSchemas(@TempDir final Path tempDir) throws IOException {

        // Given
        final String testConfigYaml = getResourceAsString("test-config.yml");
        final Path configPath = Files.writeString(Path.of(tempDir.toString(), "test-config.yml"), testConfigYaml);

        // When
        final ApplicationConfig config = ApplicationConfigLoader.loadConfigFromPath(configPath);

        // Then
        assertNotNull(config.getSubmissionDataSchemas());
        assertThat(config.getSubmissionDataSchemas().size(), is(3));
    }

    private String getResourceAsString(final String name) throws IOException {
        final ClassLoader classLoader = getClass().getClassLoader();
        final File file = new File(Objects.requireNonNull(classLoader.getResource(name)).getFile());
        return Files.readString(file.toPath());
    }
}